/*
 * Written by Jimmie Alam and Brian Birtles, 2004
 * University of Technology, Sydney
 *
 * Based on the MegaAVR port.
 *
 * Contains some code written by Vladimir Lasky from the Faculty of Engineering
 * at University of Technology, Sydney. Used by permission.
 */

#include "projdefs.h"
#include "portable.h"
#include "task.h"
#include "mcf5272.h"

/*-----------------------------------------------------------
 * Implementation of functions defined in portable.h for the MCF5272 port.
 *----------------------------------------------------------*/

/* Start tasks with interrupts enables. */
#define portFLAGS_INT_ENABLED					( ( portSTACK_TYPE ) 0x2000 )
#define portINITIAL_EXCEPTION_STACK_FRAME 		( ( portSTACK_TYPE ) \
													  EXSF_FORMAT(4) \
													| EXSF_FS(0) \
													| EXSF_VECTOR(VECTOR_TMR0) \
													| STATUS_S \
													| STATUS_I(0) )

/* Hardware constants for timer 0. */
#define portCLOCK_PRESCALER						( ( unsigned portLONG ) 131 )

/*-----------------------------------------------------------*/

/* We require the address of the pxCurrentTCB variable, but don't want to know
any details of its type. */
typedef void tskTCB;
extern volatile tskTCB * volatile pxCurrentTCB;

/*-----------------------------------------------------------*/

/* 
 * Macro to save all the general purpose registers, the save the stack pointer
 * into the TCB.  
 * 
 * The first thing we do is save the flags then disable interrupts.  This is to 
 * guard our stack against having a context switch interrupt after we have already 
 * pushed the registers onto the stack - causing the 32 registers to be on the 
 * stack twice. 
 * 
 * r1 is set to zero as the compiler expects it to be thus, however some
 * of the math routines make use of R1. 
 * 
 * The interrupts will have been disabled during the call to portSAVE_CONTEXT()
 * so we need not worry about reading/writing to the stack pointer. 
 */

#define portSAVE_CONTEXT()									\
	asm volatile (	"move.l %d0,%sp@-				\n\t"	\
					"move.l %d1,%sp@-				\n\t"	\
					"move.l %d2,%sp@-				\n\t"	\
					"move.l %d3,%sp@-				\n\t"	\
					"move.l %d4,%sp@-				\n\t"	\
					"move.l %d5,%sp@-				\n\t"	\
					"move.l %d6,%sp@-				\n\t"	\
					"move.l %d7,%sp@-				\n\t"	\
					"move.l %a0,%sp@-				\n\t"	\
					"move.l %a1,%sp@-				\n\t"	\
					"move.l %a2,%sp@-				\n\t"	\
					"move.l %a3,%sp@-				\n\t"	\
					"move.l %a4,%sp@-				\n\t"	\
					"move.l %a5,%sp@-				\n\t"	\
					"move.l %a6,%sp@-				\n\t"	\
					"move.l pxCurrentTCB,%a0		\n\t"	\
					"move.l %sp,(%a0)				\n\t" );

/* 
 * Opposite to portSAVE_CONTEXT().  Interrupts will have been disabled during
 * the context save so we can write to the stack pointer. 
 */

#define portRESTORE_CONTEXT()								\
	asm volatile (	"move.l pxCurrentTCB, %a0		\n\t"	\
					"move.l (%a0),%sp				\n\t"	\
					"move.l %sp@+,%a6				\n\t"	\
					"move.l %sp@+,%a5				\n\t"	\
					"move.l %sp@+,%a4				\n\t"	\
					"move.l %sp@+,%a3				\n\t"	\
					"move.l %sp@+,%a2				\n\t"	\
					"move.l %sp@+,%a1				\n\t"	\
					"move.l %sp@+,%a0				\n\t"	\
					"move.l %sp@+,%d7				\n\t"	\
					"move.l %sp@+,%d6				\n\t"	\
					"move.l %sp@+,%d5				\n\t"	\
					"move.l %sp@+,%d4				\n\t"	\
					"move.l %sp@+,%d3				\n\t"	\
					"move.l %sp@+,%d2				\n\t"	\
					"move.l %sp@+,%d1				\n\t"	\
					"move.l %sp@+,%d0				\n\t" ); 

/*-----------------------------------------------------------*/

/*
 * Perform hardware setup to enable ticks from timer 1, compare match A.
 */
static void prvSetupTimerInterrupt( void );

/*
 * Timer service routine
 */
void vTimerISR( void );
/*-----------------------------------------------------------*/

/* 
 * See header file for description. 
 */
portSTACK_TYPE *pxPortInitialiseStack( portSTACK_TYPE *pxTopOfStack,
										pdTASK_CODE pxCode,
										void *pvParameters )
{
	/* Place a few bytes of known values on the bottom of the stack. 
	This is just useful for debugging. */

	*pxTopOfStack = 0x11;
	pxTopOfStack--;
	*pxTopOfStack = 0x22;
	pxTopOfStack--;
	*pxTopOfStack = 0x33;
	pxTopOfStack--;

	/* Simulate how the stack would look after a call to vPortYield() generated
	 * by the compiler. */

	/*lint -e950 -e611 -e923 Lint doesn't like this much - but nothing I can do
	 * about it. */

	/* Place the parameter on the stack in the expected location. */
	*pxTopOfStack = ( portSTACK_TYPE ) pvParameters;
	pxTopOfStack--;

	/* Place dummy return address on stack. Tasks should never terminate so
	 * we can set this to anything. */
	*pxTopOfStack = ( portSTACK_TYPE ) 0;
	pxTopOfStack--;

	/* The start of the task code will be popped off the stack last, so place
	it on first. */
	*pxTopOfStack = ( portSTACK_TYPE ) pxCode;
	pxTopOfStack--;

	/* An initial exception stack frame emulating a timer interrupt as everytime
	 * we return from a context switch we use RTE instead of RTS. */
	*pxTopOfStack = portINITIAL_EXCEPTION_STACK_FRAME;
	pxTopOfStack--;

	*pxTopOfStack = ( portSTACK_TYPE ) 0x00;	/* D0 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x01;	/* D1 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x02;	/* D2 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x03;	/* D3 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x04;	/* D4 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x05;	/* D5 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x06;	/* D6 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x07;	/* D7 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x00;	/* A0 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x01;	/* A1 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x02;	/* A2 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x03;	/* A3 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x04;	/* A4 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x05;	/* A5 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x06;	/* A6 / FP */

	/*lint +e950 +e611 +e923 */

	return pxTopOfStack;
}
/*-----------------------------------------------------------*/

portSHORT sPortStartScheduler( portSHORT sUsePreemption )
{
	if (sUsePreemption)
	{
		/* Setup the hardware to generate the tick. */
		prvSetupTimerInterrupt();
	}

	/* Restore the context of the first task that is going to run. */
	portRESTORE_CONTEXT();

	/* Simulate a function call end as generated by the compiler.  We will now
	jump to the start of the task the context of which we have just restored. */
	asm volatile ( "rte" );

	/* Should not get here. */
	return pdTRUE;
}
/*-----------------------------------------------------------*/

void vPortEndScheduler( void )
{
	/* It is unlikely that the MCF5272 port will get stopped.  If required simply
	disable the tick interrupt here. */
}
/*-----------------------------------------------------------*/

/*
 * Manual context switch.  The first thing we do is save the registers so we
 * can use a naked attribute.
 */
void vPortYield( void );
void vPortYield( void )
{
	/* As the naked attribute is not supported for m68k we need to reverse the
	 * effects of the function prologue by removing the frame pointer. */
#ifndef OMIT_FRAME_POINTER
	asm volatile ( "unlk %fp" );
#endif

	portDISABLE_INTERRUPTS();
	/* Save status register but reenable interrupts */
	asm volatile (	"move.l %%d1,-8(%%sp)		\n\t"	\
					"clr.l %%d1					\n\t"	\
					"move.w %%sr,%%d1			\n\t"	\
					"ori.l %0,%%d1				\n\t"	\
					"andi.l %1,%%d1			\n\t"	\
					"move.l %%d1,%%sp@-			\n\t"	\
					"move.l -4(%%sp),%%d1		\n\t" 	\
					: /* no output */					\
					: "n" (EXSF_FORMAT(4) | EXSF_VECTOR(VECTOR_TMR0)), \
					  "n" (~STATUS_I(7)) );

	portSAVE_CONTEXT();
	vTaskSwitchContext();
	portRESTORE_CONTEXT();

	asm volatile ( "rte" );
}
/*-----------------------------------------------------------*/

/*
 * Setup timer 1.
 */
static void prvSetupTimerInterrupt( void )
{
	/*
	 * This code is based on code written by Vladimir Lasky from the Faculty of
	 * Engineering at the University of Technology, Sydney. Used by permission.
	 *
	 * This code may be indirectly based on the example code provided by David
	 * Fiddes with the gcc-m68k compiler which has been released for free use. 
	 */
	typedef struct
	{
		void (*handler[255])();
	} VectorTable;

	volatile VectorTable *vecttable = (VectorTable *)VBR;

	asm volatile (	"move.l %%a7,%%d1	\n\t"	\
					"move.l %0,%%a7		\n\t"	\
					"movec.l %%a7,%%vbr	\n\t"	\
					"move.l %%d1,%%a7	\n\t"	\
					: : "n" (VBR) : "d1" );
	
	vecttable->handler[VECTOR_TMR0] = vTimerISR;

	ICR1 = ICR1_TMR0PI | ICR1_TMR0PL(5);

	TMR0 = TMRX_RST;
	TMR0 = 0;
	TCN0 = 0;
	TER0 = TERX_REF | TERX_CAP;

	TRR0 = (portCPU_CLOCK_HZ / portTICK_RATE_HZ) / portCLOCK_PRESCALER;
	
	TMR0 = TMRX_PRESCALER(portCLOCK_PRESCALER) |
		TMRX_CE(0) |
		TMRX_ORI |
		TMRX_FRR |
		TMRX_CLK(0x1) |
		TMRX_RST;
}
/*-----------------------------------------------------------*/

/*
 * Tick ISR for preemptive scheduler.
 */
void vTimerISR( void )
{
	portDISABLE_INTERRUPTS();
/* Yes, there is also a frame pointer for an interrupt */
#ifndef OMIT_FRAME_POINTER
	asm volatile ( "unlk %fp" );
#endif
	portSAVE_CONTEXT();
	vTaskIncrementTick();
	vTaskSwitchContext();
	portRESTORE_CONTEXT();

	asm volatile ( "rte" );
}
/*-----------------------------------------------------------*/

unsigned portSHORT usPortCheckFreeStackSpace( const unsigned portCHAR *pucStackByte, unsigned portCHAR ucStackFillByte )
{
register unsigned portSHORT usCount = 0;

	while( *pucStackByte == ucStackFillByte )
	{
		pucStackByte -= portSTACK_GROWTH;
		usCount++;
	}

	usCount /= sizeof( portSTACK_TYPE );

	return usCount;
}
/*-----------------------------------------------------------*/

